from datetime import datetime, timedelta
from bson.objectid import ObjectId

from ..validator import method
from ..model.lowest import rtoken as lmodel_rtoken
from ..model.engine import rtoken as emodel_rtoken


class RtokenRepo(object):
    '''
    Provide methods to manage refresh token

    :param pymongo.MongoClien db: Instance of connection to mongdb server
    '''

    def __init__(self, db):
        self._db = db
        self._coll = db['rtoken']

    @method(lmodel_rtoken.attr_id, ObjectId, timedelta)
    def insert_one(self, rtoken_id, account_id, expires_in=timedelta(days=30)):
        '''
        Insert a refresh token

        :param string rtoken_id: Refresh token identity
        :param bson.objectid.ObjectId account_id: Account identity which is
            own refresh token
        :param string expires_in: Time span from now specify expired of refresh
            token
        '''

        expired_date = (datetime.now() + expires_in).timestamp()

        rtoken = {
            '_id': rtoken_id,
            'account_id': account_id,
            'expired': int(expired_date)
        }

        emodel_rtoken.validate_insert(rtoken)

        self._coll.insert_one(rtoken)

    @method(lmodel_rtoken.attr_id)
    def verify_and_remove(self, rtoken_id):
        '''
        Find refresh token in storage.
        If exist, get refresh token, verify and remove from storage.
        If refresh token is valid, return refresh toke, else return None

        :param str rtoken_id: Refresh token identity
        :return: Refresh token data
        :rtype: dict
        :rtype: None
        '''

        rtoken = self._coll.find_one({'_id': rtoken_id})
        if rtoken is None:
            return None

        self._coll.delete_one({'_id': rtoken_id})

        now = datetime.now().timestamp()
        if rtoken['expired'] < now:
            return None

        return rtoken



import sys
import click
import configparser
import os
import pkg_resources
from os import path
from .service import Service


root = path.realpath(path.join(path.dirname(__file__), '..'))
version = pkg_resources.get_distribution('gwisp').version


def read_config():
    config_file = path.join(root, 'config.conf')
    if not path.isfile(config_file):
        if os.name == 'posix':
            config_file = '/etc/gwisp/config.conf'
        else:
            config_file = '%SYSTEMDRIVE%\\gwisp\\config.conf'

    config = configparser.ConfigParser()
    config['DEFAULT'] = {
        'port': 9001,
        'redirect_uri': 'http://localhost:9001/oauth/callback',
        'db_url': 'mongodb://localhost/gwisp',
        'client_id': '736160870024-0cu341if9scqjfpvdsaqhutjebobjjq0.apps.'
                     'googleusercontent.com',
        'client_secret': 'hiznIzdC508X3sNOJpSfCrx7',
        'jwt_key': 'mirana-9870'
    }
    config.read([config_file])
    if 'APP' not in config:
        return config['DEFAULT']

    return config['APP']


def get_service():
    config = read_config()

    return Service(int(config['port']), config['db_url'],
                   config['client_id'], config['client_secret'],
                   config['redirect_uri'], config['jwt_key'])


@click.group()
@click.version_option(version=version)
def cli():
    pass


@cli.command(help='Register root account')
@click.option('--passwd', required=True, help='Password of root account')
@click.option('--email', required=True, help='Email of root account')
def regroot(passwd, email):
    get_service().reg_root(email, passwd)


@cli.command(help='Create new database with empty data')
def newdb():
    get_service().renew_db()


@cli.command(help='Create new database then push sample data to it')
@click.option('--yes', count=True, help='skip confirm message')
def new(yes):
    if not yes:
        confirm = input('This action will drop databse y/n: ')
        if confirm.lower() != 'y':
            sys.exit(0)

    get_service().setup(path.join(root, 'asset/sample-data'))


@cli.command(help='Start service')
def start():
    get_service().start()


@cli.command(help='Show information of configuration file')
def info():
    config = read_config()

    print('port: {:s}'.format(config['port']))
    print('db_url: {:s}'.format(config['db_url']))
    print('client_id: {:s}'.format(config['client_id']))
    print('client_secret: {:s}'.format(config['client_secret']))
    print('redirect_uri: {:s}'.format(config['redirect_uri']))



import copy
import jsonschema

from ..lowest import rtoken


#: Model use to validate insert data
insert = copy.deepcopy(rtoken.storage)


def validate_insert(item):
    '''
    Validate insert data

    :param dict: Data use to insert
    :raise jsonschema.exceptions.ValidationError: On data is invalid
    '''

    jsonschema.validate(item, insert)



import copy
import jsonschema

from ..lowest import scheduler


#: Model use to validate insert data
insert = copy.deepcopy(scheduler.storage)
insert['required'].remove('_id')
insert['properties'].pop('_id')

#: Model use to validate update data
update = copy.deepcopy(scheduler.storage)
update.pop('required')
update['properties'].pop('_id')
update['properties'].pop('name')


def validate_insert(items):
    '''
    Validate insert data

    :param object items: List or single data use to insert
    :raise jsonschema.exceptions.ValidationError: On data is invalid
    '''

    if type(items) is list:
        for item in items:
            jsonschema.validate(item, insert)
    else:
        jsonschema.validate(items, insert)


def validate_update(item):
    '''
    Validate update data

    :param object item: Data use to insert
    :raise jsonschema.exceptions.ValidationError: On data is invalid
    '''

    jsonschema.validate(item, update)



import functools
import jsonschema
from inspect import getargspec


class ArgTypeError(TypeError):
    '''
    Use to raise exception when Argument type is invalid

    :param string func_name: Name of function
    :param int arg_oder: Order of arugment of function, base on zero
    :param object arg_value: Value of argumen
    :param string req_arg_type: Name of argument type is requiredt
    '''

    def __init__(self, func_name, arg_oder, arg_value, req_arg_type):
        self.error = 'The argument {} of {}() is not a {}. It is {}'.format(
            arg_oder, func_name, req_arg_type, type(arg_value)
        )

    def __str__(self):
        return self.error


class ArgNumberError(TypeError):
    '''
    Use to raise exception when number required arguments is not equal
    with number of function arguments

    :param string fn_name: Name of fucntion
    :param int num_args: Number of arguments in function
    '''

    def __init__(self, fn_name, num_args):
        self.error = 'Required full spec for {} arguments in {}()'.format(
            num_args, fn_name
        )

    def __str__(self):
        return self.error


def validate_args(fn_name, fn_args, req_args):
    '''
    Validate arguments of function. If type of element in req_args is dict,
    supported to validate by jsonschema. If type of element in req_args is,
    type, supported to compare type

    :param string fn_name: Name of function to be check
    :param tuple fn_args: Instance of arguments in order
    :param tuple req_args: Required type of arguments in Order

    :raise gwisp.validator.ArgNumberError: On number of required arguments
        is not enough
    :raise gwisp.validator.ArgTypeError: On type of arguments in function
        is invalid
    :raise jsonschema.exceptions.ValidationError: On value of arguments
        in function invalid jsonschema
    '''

    args = enumerate(zip(fn_args, req_args))

    for arg_num, (fn_arg, req_arg) in args:
        req_arg_type = type(req_arg)

        # check function type vs required type
        if req_arg_type is type:
            if type(fn_arg) is not req_arg:
                raise ArgTypeError(fn_name, arg_num, fn_arg, req_arg)

        # check function value with jsonschema
        elif req_arg_type is dict:
            jsonschema.validate(fn_arg, req_arg)

        # do not supported other case
        else:
            raise TypeError('Does not supported type of arguments')


def method(*req_args):
    '''
    This is decorator. Use to be check type of arguments in methods

    :param req_args: List of required arguments

    :raise gwisp.validator.ArgNumberError: On number of required arguments
        is not enough
    :raise gwisp.validator.ArgTypeError: On type of arguments in function
        is invalid
    :raise jsonschema.exceptions.ValidationError: On value of arguments
        in function invalid jsonschema
    '''

    def accept_decorator(fn):
        @functools.wraps(fn)
        def decorator_wrapper(*fn_args):
            argspec = getargspec(fn)
            fn_args_full = fn_args
            if argspec.defaults is not None:
                default_num = len(argspec.args) - len(fn_args)
                if default_num > 0:
                    fn_args_full = fn_args + argspec.defaults[-default_num:]

            if len(req_args) is not (len(fn_args_full) - 1):
                raise ArgNumberError(fn.__name__, len(fn_args_full))

            validate_args(fn.__name__, fn_args_full[1:], req_args)

            return fn(*fn_args)

        return decorator_wrapper

    return accept_decorator



from os import path
from pymongo import MongoClient
from .collection import specs
from .error import CollectionIndexError


class DbClient(object):
    '''
    Provide general action of mongodb. And guaranteed that make a database
    with correct indexs

    :param str url: Url to mongodb server, include username and password
    '''
    def __init__(self, url):
        self._url = url
        self._dbname = path.basename(self._url)

    @property
    def db(self):
        '''
        :return: Instance of connection to mongodb
        :rtype: pymongo.MongoClient
        '''

        return self._db

    def connect(self):
        '''
        Connect to mongodb server
        '''

        self._mongo_client = MongoClient(self._url)
        self._db = self._mongo_client[self._dbname]

    def close(self):
        '''
        Close connection to mongodb server
        '''

        self._mongo_client.close()
        self._mongo_client = None

    def drop(self):
        '''
        Drop current database
        '''

        self._mongo_client.drop_database(self._dbname)
        self._mongo_client = None

    def check_indexs(self):
        '''
        Check indexs of database

        :raise CollectionIndexError: Indexes of collection is invalid
        '''

        # check for each collection has defined
        for spec in specs:
            db_indexes = self._db[spec.name].index_information()

            # check for each index of collection has defined
            for index_name in spec.indexes:
                if index_name not in db_indexes:
                    raise CollectionIndexError(spec.name)
                db_index = db_indexes[index_name]

                if spec.indexes[index_name]['key'] != db_index['key']:
                    raise CollectionIndexError(spec.name)
                if spec.indexes[index_name]['unique'] != db_index['unique']:
                    raise CollectionIndexError(spec.name)

    def renew(self):
        '''
        If database is early exist then drop it
        Create new database with empty collections
        '''

        self.drop()
        self.connect()

        for spec in specs:
            self._db.create_collection(spec.name)
            coll = self._db[spec.name]

            for index_name in spec.indexes.keys():
                coll.create_index(
                    spec.indexes[index_name]['key'],
                    name=index_name,
                    unique=spec.indexes[index_name]['unique']
                )

    def is_empty(self):
        '''
        Check database is empty or not

        :returns: True on database is empty
        :returns: False on database is not empty
        :rtype: bool
        '''

        return len(self._db.collection_names()) == 0



import copy

from .get_scheduler import res_body as get_sched_res_body
from ..repo import cscheduler as rmodel_csched


res_body = copy.deepcopy(get_sched_res_body)
res_body['items']['required'].extend(['_cloned', '_updated'])
res_body['items']['properties'].update({
    '_cloned': rmodel_csched.attr_cloned,
    '_updated': rmodel_csched.attr_updated
})
res = {'body': res_body}



from .ess import account_id

#: Identity
attr_id = {
    'type': 'string',
    'pattern': '^[a-zA-Z0-9]{22}$'
}

#: Account identity which is refresh token belong
attr_account_id = account_id

#: Expired date
attr_expired = {
    'type': 'integer'
}

#: Model store in database
storage = {
    'type': 'object',
    'additionalProperties': False,
    'required': ['_id', 'account_id', 'expired'],
    'properties': {
        '_id': attr_id,
        'account_id': attr_account_id,
        'expired': attr_expired
    }
}



import copy
import jsonschema

from ..lowest import scheduler as lmodel_sched


#: Model use to validate insert data
insert = copy.deepcopy(lmodel_sched.storage)
insert['required'].remove('_id')
insert['required'].remove('modified')
insert['required'].remove('num_user')
insert['properties'].pop('_id')
insert['properties'].pop('modified')
insert['properties'].pop('num_user')

#: Model use to validate insert multi data
insert_many = {
    'type': 'array',
    'items': insert
}

#: Model use to validate update data
update = copy.deepcopy(lmodel_sched.storage)
update.pop('required')
update['properties'].pop('_id')
update['properties'].pop('author')
update['properties'].pop('name')
update['properties'].pop('modified')


search = {
    'type': 'array',
    'items': {
        'type': 'object',
        'additionalProperties': False,
        'required': ['_id', 'author', 'name', 'num_user', 'modified'],
        'properties': {
            '_id': lmodel_sched.attr_id,
            'author': lmodel_sched.attr_author,
            'name': lmodel_sched.attr_name,
            'num_user': lmodel_sched.attr_num_user,
            'modified': lmodel_sched.attr_modified
        }
    }
}

find = copy.deepcopy(lmodel_sched.storage)


def validate_insert(items):
    '''
    Validate data use to insert

    :param object items: List or single of data to insert
    :raise jsonschema.exceptions.ValidationError: On data is invalid
    '''

    if type(items) is list:
        for item in items:
            jsonschema.validate(item, insert)
    else:
        jsonschema.validate(items, insert)


def validate_update(items):
    '''
    Validate data use to update

    :param object items: List or single of data use to update
    :raise jsonschema.exceptions.ValidationError: On data is invalid
    '''

    if type(items) is list:
        for item in items:
            jsonschema.validate(items, update)
    else:
        jsonschema.validate(items, update)



from falcon import HTTPNotFound, HTTPBadRequest
from bson import ObjectId

from ..repo import CSchedulerRepo
from ..util import parse_selector


class SchedulerUserCtl(object):
    '''
    Allow search schedulers on user view

    :param pymongo.MongoClient db: Instance of connection to mongodb server
    :param gwisp.Protector protector: Instance of protector
    '''

    def __init__(self, db, protector):
        self._protector = protector
        self._repo = CSchedulerRepo(db)

    def on_get(self, req, res):
        auth = self._protector.check(req)

        selector = parse_selector(req)

        res.body = self._repo.search_on_uview(auth.account_id, selector)
        if res.body is None:
            raise HTTPNotFound(description='Not found scheduler')


class SchedulerUserItemCtl(object):
    '''
    Allow query single shceduler on user view point

    :param pymongo.MongoClient db: Instance of connection to mongodb server
    :param gwisp.Protector protector: Instance of protector
    '''

    def __init__(self, db, protector):
        self._protector = protector
        self._repo = CSchedulerRepo(db)

    def on_get(self, req, res, id):
        auth = self._protector.check(req)

        try:
            sched_id = ObjectId(id)
        except Exception:
            raise HTTPBadRequest(
                title='400 Bad Request',
                description='Identity is invalid'
            )

        res.body = self._repo.find_by_id_on_uview(auth.account_id, sched_id)
        if res.body is None:
            raise HTTPNotFound(
                title='404 Not Found',
                description='Not found scheduler'
            )



#: Object identity
object_id = {}

#: Account identity
account_id = object_id

#: Account name
account_name = {
    'type': 'string',
    'pattern': '^[a-z0-9-]{1,64}$'
}

#: Account header, use as references
account_header = {
    'type': 'object',
    'additionalProperties': False,
    'required': ['_id', 'name'],
    'additionalProperties': False,
    'properties': {
        '_id': account_id,
        'name': account_name
    }
}

#: Email
email = {
    'type': 'string',
    'pattern': '^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$'
}

#: URL
url = {
    'type': 'string',
    'pattern': '^(https?:\/\/)?([\da-z\.-]+)\.([a-z\.]{2,6})([\/\w \.-]*)*\/?$'
}

#: Base 64 string
base_64 = {
    'type': 'string',
    'pattern': '^[a-zA-Z0-9+/]+={0,2}$'
}



import copy


from ..lowest import scheduler
from ..repo import cscheduler as rmodel_sched


res_body = copy.deepcopy(scheduler.storage)
res_body['required'].extend(['_updated'])
res_body['properties'].update({
    '_updated': rmodel_sched.attr_updated
})
res = {'body': res_body}



from falcon import HTTPNotFound, HTTPInternalServerError, HTTPBadRequest, \
                   HTTP_204
from bson import ObjectId

from ..repo import SchedulerRepo, CSchedulerRepo


class MeTargetCtl(object):
    def __init__(self, db, protector):
        self._sched_repo = SchedulerRepo(db)
        self._protector = protector

    def on_get(self, req, res):
        auth = self._protector.check(req, iacc=True)

        if 'scheduler' not in auth.account:
            raise HTTPNotFound(description='Not found target scheduler')

        if 'scheduler' not in auth.account:
            raise HTTPNotFound(description='No scheduler is using')
        target = auth.account['scheduler']

        real_item = self._sched_repo.find_by_id(target['_id'])
        if real_item is None:
            raise HTTPInternalServerError(
                description='Not found scheduler'
            )

        if real_item['modified'] > target['_cloned_date']:
            target['_updated'] = False
        else:
            target['_updated'] = True

        # remove data
        del target['_cloned_date']

        res.body = target


class MeTargetItemCtl(object):
    def __init__(self, db, protector):
        self._protector = protector
        self._csched_repo = CSchedulerRepo(db)

    def on_put(self, req, res, id):
        auth = self._protector.check(req)

        try:
            sched_id = ObjectId(id)
        except Exception as e:
            raise HTTPBadRequest(
                title='400 Bad Request',
                description='Identity is invalid'
            )

        self._csched_repo.use_scheduler(auth.account_id, sched_id)

        res.status = HTTP_204



from falcon import HTTP_204, HTTP_404, \
                   HTTPBadRequest, HTTPNotFound, HTTPForbidden
from bson.objectid import ObjectId

from ..error import NotExistError
from ..model.front.post_scheduler import validate_post
from ..model.front.put_scheduler import validate_put
from ..repo import SchedulerRepo
from ..util import parse_selector


class SchedulerCtl(object):
    '''
    Scheduler resource

    :param pymongo.MongoClient db: Instance of mongo client
    :param gwisp.Protector protector: Instance of protector
    '''

    def __init__(self, db, protector):
        self._protector = protector
        self._repo = SchedulerRepo(db)

    def on_get(self, req, res):
        '''
        Search scheduler

        Request: Query parameters as selector

        Response: List of scheduler match with selector
        '''

        res.body = self._repo.search(parse_selector(req))

        if res.body is None:
            res.status = HTTP_404

    def on_post(self, req, res):
        '''
        Create new scheduler

        Request: Scheduler in body, access token in header

        Response: Relative location of new scheduler in header, field
            Location
        '''

        if 'body' not in req.context:
            raise HTTPBadRequest('400 Bad Request', 'Body is empty')
        item = req.context['body']
        validate_post(item)

        auth = self._protector.check(req, iacc=True)

        # padding data
        item['author'] = {
            '_id': auth.account_id,
            'name': auth.account['name']
        }

        # save scheduler to storage
        id = self._repo.insert_one(item)

        res.status = HTTP_204
        res.set_header('Location', '/scheduler/{}'.format(str(id)))


class SchedulerItemCtl(object):
    '''
    Scheduler resource

    :param pymongo.MongoClient db: Instance of mongo client
    :param gwisp.Protector protector: Instance of protector
    '''

    def __init__(self, db, protector):
        self._repo = SchedulerRepo(db)
        self._protector = protector

    def on_get(self, req, res, id):
        '''
        Query single scheduler

        Request: Identify of scheduler in request url

        Response: Scheduler in body
        '''

        sched_id = None
        try:
            sched_id = ObjectId(id)
        except Exception as e:
            raise HTTPBadRequest(
                title='400 HTTPBadRequest',
                description='id is invalid'
            )

        res.body = self._repo.find_by_id(sched_id)

        if res.body is None:
            raise HTTPNotFound(
                title='404 Not Found',
                description='Not found scheduler'
            )

    def on_put(self, req, res, id):
        '''
        Update scheduler

        Request:
            - Identify of scheduler in request url
            - Scheduler content in body
            - Authorization header

        Response: No Content
        '''

        sched_id = None
        try:
            sched_id = ObjectId(id)
        except Exception as e:
            raise HTTPBadRequest(
                title='400 HTTPBadRequest',
                description='id is invalid'
            )

        auth = self._protector.check(req)
        old_item = self._repo.find_by_id(sched_id)
        if old_item is None:
            raise NotExistError('Scheduler is not exist')
        if old_item['author']['_id'] != auth.account_id:
            raise HTTPForbidden(
                title='403 Forbidden',
                description='Can not update item of other account'
            )

        if 'body' not in req.context:
            raise HTTPBadRequest('400 Bad Request', 'Invalid body')
        item = req.context['body']
        validate_put(item)

        # put data to storage
        self._repo.update_one(ObjectId(id), item)

        # result is no content
        res.status = HTTP_204

    def on_delete(self, req, res, id):
        '''
        Remove scheduler

        Request: Identify of scheduler in request url
        '''

        # authenticate
        auth = self._protector.check(req)

        # ensure item is exist
        item = self._repo.find_by_id(ObjectId(id))
        if item is None:
            raise HTTPNotFound(description='Scheduler is not exist')

        # ensure only owner have permission to delete item
        if item['author']['_id'] != auth.account_id:
            raise HTTPForbidden(
                title='403 Forbidden',
                description='Can not delete item of other account'
            )

        # remove item from storage
        self._repo.delete_one(ObjectId(id))

        # result is no content
        res.status = HTTP_204



from falcon import HTTPBadRequest, HTTP_204

from ..error import AuthenticateError
from ..repo import AccountRepo


class OauthCtl(object):

    '''
    Implement oauth-2 endpoint by essential supported

    :param gwisp.protector protector: Instance of protector
    '''

    def __init__(self, protector):
        self._protector = protector

    def on_post(self, req, res):
        params = req.context['body']

        if 'grant_type' not in params:
            raise HTTPBadRequest(
                title='400 Bad Request',
                description='Not support grant_type'
            )

        if params['grant_type'] == 'password':
            self._grant_passwd(req, res)
        elif params['grant_type'] == 'refresh_token':
            self._grant_rtoken(req, res)
        else:
            raise HTTPBadRequest(
                title='400 Bad Request',
                description='Not support grant_type'
            )

    def _grant_passwd(self, req, res):
        params = req.context['body']

        if 'username' not in params:
            raise AuthenticateError('Username is not set')
        if 'password' not in params:
            raise AuthenticateError('Passowrd is not set')

        res.body = self._protector.login_by_passwd(
            params['username'],
            params['password']
        )

    def _grant_rtoken(self, req, res):
        params = req.context['body']

        if 'refresh_token' not in params:
            raise AuthenticateError('Refresh token is not set')

        res.body = self._protector.refresh_token(params['refresh_token'])


class LoginUrlCtl(object):
    '''
    Information about how to login

    :param gwisp.Protector protector: Instance of protector
    '''

    def __init__(self, protector):
        self._protector = protector

    def on_get(self, req, res):
        '''
        Response: Login url
        '''

        res.body = {
            'url': self._protector.login_url
        }


class LoginCtl(object):
    '''
    Support login by google oauth-2

    :param gwisp.Protector protector: Instance of protector
    '''

    def __init__(self, protector):
        self._protector = protector

    def on_get(self, req, res):
        '''
        Request: Login code from google as query param

        Response: Token string
        '''

        res.body = self._protector.login_by_code(req.params['code'])


class MeCtl(object):
    '''
    Account information

    :param pymongo.MongoClient db: Instance of mongo client
    :param gwisp.Protector protector: Instance of protector
    '''

    def __init__(self, db, protector):
        self._protector = protector
        self._acc_repo = AccountRepo(db)

    def on_get(self, req, res):
        '''
        Request: Token in header

        Response: Account information
        '''

        info = self._protector.check(req, True)
        del info.account['subject']
        res.body = info.account

    def on_put(self, req, res):
        '''
        Request: Token in header and content in body

        Response: No Content
        '''

        auth = self._protector.check(req)

        self._acc_repo.update_one(auth.account_id, req.context['body'])

        res.status = HTTP_204



import copy

from ..lowest import scheduler as lmodel_sched
from ..repo import cscheduler as rmodel_csched


res_body = copy.deepcopy(lmodel_sched.storage)
res_body['required'].extend(['_updated'])
res_body['properties'].update({
    '_updated': rmodel_csched.attr_updated
})
res = {'body': res_body}



import jsonschema

from ..lowest import password as password_lmodel


def validate_insert(account_id, password_hash):
    '''
    Validate data use to insert

    :param bson.ObjectId account_id: Account identity corresponding with
        password
    :param string password_hash: Password after hash
    :raise jsonschema.exceptions.ValidationError: On data is invalid
    '''

    jsonschema.validate(account_id, password_lmodel.attr_id)
    jsonschema.validate(password_hash, password_lmodel.attr_password_hash)


def validate_update(account_id, password_hash):
    '''
    Validate data use to update

    :param bson.ObjectId account_id: Account identity corresponding with
        password
    :param string password_hash: Password after hash
    :raise jsonschema.exceptions.ValidationError: On data is invalid
    '''

    jsonschema.validate(account_id, password_lmodel.attr_id)
    jsonschema.validate(password_hash, password_lmodel.attr_password_hash)



import copy
import jsonschema

from .ess import account_id, account_name, email, url
from .scheduler import storage as sched_storage


#: Identity
attr_id = account_id

#: Name
attr_name = account_name

#: Email
attr_email = email

#: Identity of oauth account
attr_subject = {
    'type': 'string'
}

#: Groups which is account belong
attr_groups = {
    'type': 'array',
    'items': {
        'type': 'string',
        'pattern': '^[a-z0-9]{1,8}$'
    }
}

#: Scheduler using by account
attr_scheduler = copy.deepcopy(sched_storage)
attr_scheduler['required'].extend(['_cloned_date'])
attr_scheduler['properties'].update({
    '_cloned_date': {'type': 'integer'}
})

#: Language
attr_language = {
    'type': 'string',
    'pattern': '^[a-z]{2}$'
}

#: Avatar link
attr_avatar = url

#: Model store in database
storage = {
    'type': 'object',
    'additionalProperties': False,
    'required': [
        '_id', 'name', 'email',
        'groups', 'language', 'avatar', 'scheduler'
    ],
    'properties': {
        '_id': attr_id,
        'name': attr_name,
        'groups': attr_groups,
        'language': attr_language,
        'avatar': attr_avatar,
        'email': attr_email,
        'subject': attr_subject,
        'scheduler': attr_scheduler
    }
}


def validate_groups(groups):
    '''
    Validate groups of account

    :param object item: Account object
    :raise jsonschema.exceptions.ValidationError: On data is invalid
    '''

    jsonschema.validate(groups, attr_groups)



import json
import glob
from os import path
from gwisp.repo import SchedulerRepo, AccountRepo


class DbTool(object):
    '''
    Controll install sample data into database

    :param pymongo.MongoClient db: Instance of connection to mongodb server
    :param str asset_dir: Path to static asset directory. That's directory
        must contains file specifications. See :ref:`api_asset_dir`
    '''
    def __init__(self, db, asset_dir, account):
        self._db = db                   #: Instance of mongodb client
        self._asset_dir = asset_dir     #: Path to asset directory

        self._sched_repo = SchedulerRepo(db)    #: Scheduler repository
        self._acc_repo = AccountRepo(db)        #: Account repository

        #: Real account identity
        self._account = account
        self._acc_id = self._acc_repo.insert_one(account)

    def install(self):
        '''
        Install sample data into database

        :return: Number of data files has been readed
        :rtype: int
        '''

        return self.install_scheduler()

    def install_scheduler(self):
        '''
        Install sample scheduler into database

        :return: Number of data files has been readed
        :rtype: int
        '''

        files = glob.glob(path.join(self._asset_dir, 'scheduler-*.json'))
        items = []
        nfiles = 0

        for file in files:
            nfiles += 1

            with open(file) as f:
                item = json.load(f)
                item['author'] = {
                    '_id': str(self._acc_id),
                    'name': self._account['name']
                }
                items.append(item)

            if len(items) == 128:
                self._sched_repo.insert_many(items)
                items = []

        if len(items) > 0:
                self._sched_repo.insert_many(items)

        return nfiles



import copy
import jsonschema

from ..repo import scheduler as rmodel_sched


req_body = copy.deepcopy(rmodel_sched.insert)
req_body['required'].remove('author')
req_body['properties'].pop('author')
req = {'body': req_body}

res_header = {
    'type': 'object',
    'required': ['Location'],
    'properties': {
        'Location': {
            'type': 'string',
            'pattern': '^\/scheduler\/[a-fA-F0-9]{24}$'
        }
    }
}
res = {'header': res_header}


def validate_post(item):
    jsonschema.validate(item, req_body)



import copy

from .get_scheduler_id import res_body as get_sched_res_body
from ..repo import cscheduler as rmodel_sched


res_body = copy.deepcopy(get_sched_res_body)
res_body['required'].extend(['_cloned', '_updated'])
res_body['properties'].update({
    '_cloned': rmodel_sched.attr_cloned,
    '_updated': rmodel_sched.attr_updated
})
res = {'body': res_body}



from falcon import HTTPNotFound, HTTP_204, HTTPBadRequest
from bson import ObjectId

from ..repo import SchedulerRepo, CSchedulerRepo
from ..util import parse_selector


class MeSchedulerCtl(object):
    def __init__(self, db, protector):
        self._repo = CSchedulerRepo(db)
        self._protector = protector

    def on_get(self, req, res):
        auth = self._protector.check(req)
        selector = parse_selector(req)

        res.body = self._repo.search_on_cview(auth.account_id, selector)

        if len(res.body) == 0:
            raise HTTPNotFound(description='Not found your scheduler')

    def on_put(self, req, res):
        auth = self._protector.check(req)

        self._repo.update_all_cloned(auth.account_id)

        res.status = HTTP_204


class MeSchedulerItemCtl(object):
    def __init__(self, db, protector):
        self._sched_repo = SchedulerRepo(db)
        self._repo = CSchedulerRepo(db)
        self._protector = protector

    def on_get(self, req, res, id):
        auth = self._protector.check(req)

        try:
            sched_id = ObjectId(id)
        except Exception:
            raise HTTPBadRequest(
                title='400 Bad Request',
                description='Identity is invalid'
            )

        res.body = self._repo.find_by_id_on_cview(
            auth.account_id, sched_id
        )
        if res.body is None:
            raise HTTPNotFound(description='Not found cloned scheduler')

    def on_put(self, req, res, id):
        auth = self._protector.check(req)

        sched_id = None
        try:
            sched_id = ObjectId(id)
        except Exception:
            raise HTTPBadRequest(
                title='400 Bad Request',
                description='Identity is invalid'
            )

        self._repo.update_one_cloned(auth.account_id, sched_id)

        res.status = HTTP_204

    def on_post(self, req, res, id):
        auth = self._protector.check(req)

        sched_id = None
        try:
            sched_id = ObjectId(id)
        except Exception:
            raise HTTPBadRequest(
                title='400 Bad Request',
                description='Identity is invalid'
            )

        self._repo.clone_one(auth.account_id, sched_id)

        res.status = HTTP_204
        res.append_header(
            'Location', '/me/scheduler/{}'.format(sched_id)
        )

    def on_delete(self, req, res, id):
        auth = self._protector.check(req)

        sched_id = None
        try:
            sched_id = ObjectId(id)
        except Exception:
            raise HTTPBadRequest(
                title='400 Bad Request',
                description='Identity is invalid'
            )

        self._repo.remove_one_cloned(auth.account_id, sched_id)

        res.status = HTTP_204



import falcon
from pymongo.errors import DuplicateKeyError
from jwt.exceptions import DecodeError, ExpiredSignatureError
from jsonschema.exceptions import ValidationError

from .handler import dupkey_handle, jwt_decode_handle, \
                     unauthorized_handle, forbidden_handle, jwt_sign_handle, \
                     validation_handle, not_exist_handle
from .error import AuthenticateError, AuthorizeError, NotExistError
from .dbclient import DbClient
from .dbtool import DbTool
from .middleware import RequireJSON, JsonTranslator, FormTranslator
from .ctl import OauthCtl, \
                 RootCtl, LoginUrlCtl, LoginCtl, MeCtl, \
                 SchedulerItemCtl, SchedulerCtl, \
                 SchedulerUserCtl, SchedulerUserItemCtl, \
                 SchedulerMeCtl, \
                 MeTargetCtl, MeTargetItemCtl, \
                 MeSchedulerCtl, MeSchedulerItemCtl
from .protector import Protector


class Worker(object):
    '''
    Create WSGI application with gwisp http interface.
    No real serve anythings. After get application from ``start()``,
    use that application to serve though ``guicorn`` server

    :param str db_url: Url to mongo database server
    :param str client_id: Google client identity, use for oauth-2
    :param str client_secret: Google client secret, use for oauth-2
    :param str redirect_uri: Callback url after confirm google login
    :param str jwt_key: Secret words use to sign token of service
    '''
    def __init__(self, db_url,
                 client_id, client_secret, redirect_uri, jwt_key):
        self._db_url = db_url
        self._client_id = client_id
        self._client_secret = client_secret
        self._redirect_uri = redirect_uri
        self._jwt_key = jwt_key

        # Connect to mongodb server
        # All of task must use only this connection
        self._dbc = DbClient(self._db_url)
        self._dbc.connect()

        # Protector use to authenticate and authorize
        self._protector = Protector(
            self._dbc.db,
            self._client_id,
            self._client_secret,
            [
                'https://www.googleapis.com/auth/plus.me',
                'profile',
                'email'
            ],
            self._redirect_uri,
            self._jwt_key
        )

    def reg_root(self, email, passwd):
        '''
        Register root account. Name is gwisp

        :param str email: Email of root account
        :param str passwd: Password in plain text
        '''

        account = {
            'name': 'gwisp',
            'email': email,
            'subject': '0',
            'language': 'en',
            'groups': ['root'],
            'avatar': 'http://static.gwisp/avatar/gwisp.jpg'
        }

        self._protector.create_account(account, passwd)

    def start(self):
        '''
        Create an WSGI application

        :return: WSGI application
        :rtype: falcon.API
        '''

        # Create collection if database is empty
        if self._dbc.is_empty() is True:
            self._dbc.renew()

        # Check indexes of collection in database
        self._dbc.check_indexs()

        # Shortcut for connection to database
        db = self._dbc.db

        # Http application
        app = falcon.API(middleware=[
            RequireJSON(),
            JsonTranslator(),
            FormTranslator()
        ])

        # Error handle
        app.add_error_handler(ValidationError, validation_handle)
        app.add_error_handler(DuplicateKeyError, dupkey_handle)
        app.add_error_handler(DecodeError, jwt_decode_handle)
        app.add_error_handler(ExpiredSignatureError, jwt_sign_handle)
        app.add_error_handler(AuthenticateError, unauthorized_handle)
        app.add_error_handler(AuthorizeError, forbidden_handle)
        app.add_error_handler(NotExistError, not_exist_handle)

        # Controllers
        app.add_route('/', RootCtl(db))

        app.add_route('/oauth', OauthCtl(self._protector))
        app.add_route('/oauth/url', LoginUrlCtl(self._protector))
        app.add_route('/oauth/callback', LoginCtl(self._protector))

        app.add_route('/me', MeCtl(db, self._protector))

        app.add_route('/scheduler', SchedulerCtl(db, self._protector))
        app.add_route('/scheduler/{id}', SchedulerItemCtl(db, self._protector))

        app.add_route('/scheduler/user', SchedulerUserCtl(db, self._protector))
        app.add_route(
            '/scheduler/user/{id}',
            SchedulerUserItemCtl(db, self._protector)
        )

        app.add_route('/scheduler/me', SchedulerMeCtl(db, self._protector))

        app.add_route('/me/target', MeTargetCtl(db, self._protector))
        app.add_route('/me/target/{id}', MeTargetItemCtl(db, self._protector))
        app.add_route('/me/scheduler', MeSchedulerCtl(db, self._protector))
        app.add_route(
            '/me/scheduler/{id}',
            MeSchedulerItemCtl(db, self._protector)
        )

        return app

    def stop(self):
        '''
        Close connection to database, free other resource. If want to
        run new worker, create an new instance instead of reuse old worker
        after stop
        '''

        self._dbc.close()
        self._dbc = None
        self._protector = None

    def renew_db(self):
        '''
        Ensure database have correct index and empty data
        '''

        self._dbc.renew()

    def setup(self, asset_dir):
        '''
        Drop database if it is exist. Create new database with correct
        indexs. Install sample data

        :param str asset_dir: Path to asset directory.
            See :ref:`api_asset_dir`
        '''

        self._dbc.renew()

        account = {
            'email': 'kevin.leptons@gmail.com',
            'name': 'gwisp',
            'subject': '0',
            'language': 'en',
            'groups': ['root'],
            'avatar': 'http://static.gwisp/avatar/root.jpg'
        }

        dbtool = DbTool(self._dbc.db, asset_dir, account)
        dbtool.install()



from bson.objectid import ObjectId

from ..error import NotExistError
from ..validator import method
from ..model.repo import account as rmodel_account
from ..model.engine import account as emodel_account
from ..model.lowest import account as lmodel_account


verify_fields = {
    '_id': 0,
    'subject': 0,
    'name': 0,
    'groups': 0,
    'scheduler': 0,
    'schedulers': 0,
    'language': 0
}


class AccountRepo(object):
    '''
    Provide method to query account information in database

    :param pymongo.MongoClient db: Instance of connection to mongodb server
    '''
    def __init__(self, db):
        self._coll = db['account']
        self._sched_coll = db['scheduler']

    @method(rmodel_account.insert)
    def insert_one(self, data):
        '''
        Insert one account into database

        :param dict data: Account data
        :return: Identity of new account
        :rtype: bson.objectid.ObjectId
        :raise jsonschema.exceptions.ValidationError: On input data is invalid
        '''

        # padding data
        if 'groups' not in data:
            data['groups'] = []

        emodel_account.validate_insert(data)

        return self._coll.insert_one(data).inserted_id

    @method(ObjectId)
    def find_by_id(self, id):
        '''
        Find one account by identity

        :param bson.objectid.ObjectId id: Identity of account
        :return: Account data on found
        :return: None on not found
        :rtype: dict
        '''

        return self._coll.find_one({'_id': id})

    @method(str)
    def find_by_name(self, name):
        '''
        Find one account by name

        :param str name: Account name
        :return: Account data on found
        :return: None on not found
        :rtype: dict
        '''

        return self._coll.find_one({'name': name})

    @method(str)
    def find_by_sub(self, sub):
        '''
        Find account by subject. Subject is identity of goole account

        :param str sub: Subject to find
        :return: Account data on found
        :return: None on not found
        :rtype: dict
        '''
        return self._coll.find_one({'subject': sub})

    @method(ObjectId)
    def is_exist(self, id):
        '''
        Verify one account identity is exist

        :param bson.objectid.ObjectId id: Account identity
        :return: True on account is exist
        :return: False on account is not exist
        :rtype: bool
        '''

        return self._coll.find_one({'_id': id}, verify_fields) is not None

    @method(ObjectId, rmodel_account.update)
    def update_one(self, id, data):
        '''
        Update one account

        :param bson.objectid.ObjectId id: Account identity
        :param dict data: Data use to update
        :raise jsonschema.exceptions.ValidationError: On input data is invalid
        '''

        result = self._coll.update_one({'_id': id}, {'$set': data})

        if result.matched_count == 0:
            raise NotExistError('Account is not exists to update')

    @method(ObjectId, lmodel_account.attr_groups)
    def add_to_groups(self, acc_id, groups):
        '''
        Add account to groups. If group is early exists, do nothing

        :param bson.objectid.ObjectId acc_id: Account identity will be
            add to groups
        :param list groups: List of groups add to account
        :raise jsonschema.exceptions.ValidationError: On input data is invalid
        '''

        patch = {
            '$addToSet': {'groups': {'$each': groups}}
        }

        self._coll.update_one({'_id': acc_id}, patch)

    @method(ObjectId, lmodel_account.attr_groups)
    def remove_from_groups(self, acc_id, groups):
        '''
        Remove account from groups. If group is not exists,  do nothing

        :param bson.objectid.ObjectId acc_id: Account identity will be
            remove from groups
        :param list groups: Groups which account will be remove from
        :raise jsonschema.exceptions.ValidationError: On input data is invalid
        '''

        patch = {
            '$pull': {'groups': {'$in': groups}}
        }

        self._coll.update_one({'_id': acc_id}, patch)

    def count(self):
        '''
        Count number of accounts in collection

        :return: Number of accounts
        :rtype: int
        '''

        return self._coll.count()



class CollectionIndexError(Exception):
    '''
    Use to raise error if indexes of collection is invalid

    :param str message: Custom message related with error
    '''

    def __init__(self, message):
        self.message = message


class NotExistError(Exception):
    '''
    Use to raise error if resource is not exist

    :param str message: Custom message related with error
    '''

    def __init__(self, message):
        self.message = message


class EarlyExistError(Exception):
    '''
    Use to raise error if resource is early exist, and can not
    create same resource because resource is identitfy

    :param str message: Custom message related with error
    '''

    def __init__(self, message):
        self.message = message


class NotAllowError(Exception):
    '''
    Use to raise error if program perform update identify field of resource

    :param str message: Custom message related with error
    '''

    def __init__(self, message):
        self.message = message


class AuthenticateError(Exception):
    '''
    Use to raise error if program can not authenticate user

    :param str message: Custom message related with error
    '''

    def __init__(self, message):
        self.message = message


class AuthorizeError(Exception):
    '''
    Use to raise error if account have not enough permission to
    perform an operation

    :param str message: Custom message related with error
    '''

    def __init__(self, message):
        self.message = message



import copy
import jsonschema

from ..lowest import account


#: Model use to validate insert data
insert = copy.deepcopy(account.storage)
insert['required'].remove('_id')
insert['required'].remove('scheduler')
insert['properties'].pop('_id')


#: Model use to validate update data
update = copy.deepcopy(account.storage)
update.pop('required')
update['properties'].pop('_id')
update['properties'].pop('name')
update['properties'].pop('email')
update['properties'].pop('subject')


def validate_insert(item):
    '''
    Validate insert data

    :param dict item: Data use to insert account
    :raise jsonschema.exceptions.ValidationError: On data is invalid
    '''

    jsonschema.validate(item, insert)


def validate_update(item):
    '''
    Validate update data

    :param dict: Data use to update account
    :raise jsonschema.exceptions.ValidationError: On data is invalid
    '''

    jsonschema.validate(item, update)



#: Response result from invoking token
token = {
    'type': 'object',
    'additionalProperties': False,
    'required': [
        'access_token', 'token_type', 'expires_in', 'refresh_token'
    ],
    'properties': {
        'access_token': {'type': 'string'},
        'token_type': {'type': 'string', 'pattern': '^Bearer$'},
        'expires_in': {'type': 'integer'},
        'refresh_token': {'type': 'string'}
    }
}



from .ess import object_id, account_header


#: Identity
attr_id = object_id

#: Author
attr_author = account_header

#: Name
attr_name = {
    'type': 'string',
    'pattern': '^[a-z0-9 -.]{1,64}$'
}

#: Tags
attr_tags = {
    'type': 'array',
    'items': {
        'type': 'string',
        'pattern': '^[a-z0-9]{1,16}$'
    }
}

#: Single task
attr_tasks_elem = {
    'type': 'object',
    'additionalProperties': False,
    'required': ['start', 'action'],
    'properties': {
        'start': {
            'type': 'string',
            'pattern': '^([0-1][0-9]|2[0-3]):[0-5][0-9]$'
        },
        'action': {
            'type': 'string',
            'pattern': '^[a-zA-Z0-9 ]{1,32}$'
        }
    }
}

#: List of tasks
attr_tasks = {
    'type': 'array',
    'items': attr_tasks_elem
}

#: Number of user cloned scheduler
attr_num_user = {
    'type': 'integer'
}

#: Last modified date
attr_modified = {
    'type': 'integer'
}

#: Notes
attr_notes = {
    'type': 'string',
    'pattern': '^(.|\n){0,1024}$'
}

#: Model store in database
storage = {
    'type': 'object',
    'additionalProperties': False,
    'required': [
        '_id', 'author', 'name',
        'tags', 'num_user', 'modified',
        'tasks'
    ],
    'properties': {
        '_id': attr_id,
        'author': attr_author,
        'name': attr_name,
        'tags': attr_tags,
        'num_user': attr_num_user,
        'modified': attr_modified,
        'tasks': attr_tasks,
        'notes': attr_notes
    }
}



import httplib2
import hashlib
import random

import jwt
import datetime
from bson.objectid import ObjectId
from oauth2client.client import OAuth2WebServerFlow
from apiclient import discovery
from string import ascii_lowercase, ascii_uppercase, digits

from .error import AuthenticateError, AuthorizeError
from .repo import PasswordRepo, AccountRepo, RtokenRepo
from .util import AuthInfo


class Protector(object):
    '''
    Control authenticating and authorizing

    :param pymongo.MongoClient db: Instance of connection to mongo server
    :param string client_id: Google oauth-2 client identify
    :param string client_secret: Google oauth-2 client secret
    :param list scope: Google oauth-2 scope
    :param string redirect_uri: Oauth-2 callback uri
    :param string jwt_key: Secret word use to sign token
    '''

    rtoken_chars = ascii_uppercase + ascii_lowercase + digits

    def __init__(self, db, client_id, client_secret, scope, redirect_uri,
                 jwt_key):
        self._db = db
        self._client_id = client_id
        self._client_secret = client_secret
        self._scope = scope
        self._redirect_uri = redirect_uri
        self._jwt_key = jwt_key

        self._acc_repo = AccountRepo(db)
        self._passwd_repo = PasswordRepo(db)
        self._rtoken_repo = RtokenRepo(db)

    def create_account(self, account, password):
        '''
        Create an account with password

        :param dict account: Account information
        :param str password: Password to login
        :return: Account identity
        :rtype: bson.ObjectId
        '''

        acc_id = self._acc_repo.insert_one(account)

        passwd_hash = hashlib.sha224(password.encode('utf-8')).hexdigest()

        self._passwd_repo.insert_one(acc_id, passwd_hash)

        return acc_id

    @property
    def login_url(self):
        '''
        Get oauth-2 login url. Use browser and url to login by google oauth-2
        service

        :return: Url use to login
        :rtype: str
        '''

        oauth = self._gen_oauth2()
        return oauth.step1_get_authorize_url()

    def login_by_passwd(self, name, password):
        '''
        Login by account name and password

        :param str name: Name of account
        :param str password: Password
        :return: Token string
        :rtype: str
        '''

        account = self._acc_repo.find_by_name(name)
        if account is None:
            raise AuthenticateError('Account is not exist')

        passwd_hash = hashlib.sha224(password.encode('utf-8')).hexdigest()
        match = self._passwd_repo.is_exist(account['_id'], passwd_hash)
        if match is False:
            raise AuthenticateError('Password is incorrect')

        return self._gen_token(account['_id'])

    def login_by_code(self, code):
        '''
        Login by code from google. If google account has not early login,
        create an account

        :param str code: Code what is received from google oauth-2 service
        :return: Token what is to authenticate and authorize
        :rtype: string
        '''

        # exchange code to tokens
        oauth = self._gen_oauth2()
        credentials = oauth.step2_exchange(code)

        # use token to get google account information
        http = credentials.authorize(httplib2.Http())
        service = discovery.build('plus', 'v1')
        req = service.people().get(userId='me')
        me = req.execute(http=http)

        # verify google account is early register in system
        # if not, register first
        sub = credentials.id_token['sub']
        acc = self._acc_repo.find_by_sub(sub)
        if acc is None:
            acc = {
                'subject': sub,
                'email': credentials.id_token['email'],
                'name': me['displayName'],
                'language': me['language'],
                'avatar': me['image']['url']
            }
            self._acc_repo.insert_one(acc)

        # system token
        return self._gen_token(acc['_id'])

    def refresh_token(self, refresh_token):
        old_rtoken = self._rtoken_repo.verify_and_remove(refresh_token)
        if old_rtoken is None:
            raise AuthenticateError('Refresh token is invalid')

        return self._gen_token(old_rtoken['account_id'])

    def check(self, req, iacc=False, group=None):
        '''
        Authenticate and authorize if **group** parameter is specify

        :param falcon.Request req: Contains information about http request
        :param boolean iacc: Decide include account information
        :param string group: Group what is expected account in
        :return: Information after authenticate and authorize
        :rtype: gwisp.AuthInfo
        '''

        if 'AUTHORIZATION' not in req.headers:
            raise AuthenticateError('Authorization header is not set')

        bearer = req.headers['AUTHORIZATION'].split(' ')
        if len(bearer) != 2:
            raise AuthenticateError('Authorization header is invalid')
        if bearer[0] != 'Bearer':
            raise AuthenticateError('Only support Bearer authenticate')

        token = self.authenticate(bearer[1])
        acc_id = ObjectId(token['sub'])
        acc = None

        if iacc is True:
            acc = self._acc_repo.find_by_id(acc_id)
            if acc is None:
                raise AuthenticateError('Can not identity user')

        if group is not None:
            if iacc is False:
                acc = self._acc_repo.find_by_id(acc_id)
            if acc is None:
                raise AuthenticateError('Can no identity user')
            if group not in acc['groups']:
                raise AuthorizeError('Invalid permission')

        return AuthInfo(acc_id, token, acc)

    def authenticate(self, token):
        '''
        Identify account by decode token

        :param string token: Token to decode
        :return: Token has decoded
        :rtype: dict
        '''

        return jwt.decode(token, self._jwt_key)

    def _gen_oauth2(self):
        '''
        Create an oauth-2 web server flow object

        :return: Instance of OAuth2WebServerFlow
        :rtype: oauth2client.client.OAuth2WebServerFlow
        '''

        return OAuth2WebServerFlow(
            client_id=self._client_id,
            client_secret=self._client_secret,
            scope=self._scope,
            redirect_uri=self._redirect_uri)

    def _gen_token(self, acc_id):
        '''
        Create token string

        :param bson.objectid.ObjectId acc_id: Identity of account
        :return: Token string
        :rtype: str
        '''

        expires_in = 3600
        token_data = {
            'sub': str(acc_id),
            'exp': datetime.datetime.now().timestamp() + expires_in
        }
        access_token = jwt.encode(
            token_data,
            self._jwt_key,
            algorithm='HS512'
        )
        rtoken = self._gen_rtoken()
        self._rtoken_repo.insert_one(rtoken, acc_id)

        return {
            'token_type': 'Bearer',
            'access_token': access_token.decode(),
            'expires_in': expires_in,
            'refresh_token': rtoken
        }

    def _gen_rtoken(self):
        return ''.join(random.choice(self.rtoken_chars) for i in range(22))



import jsonschema

from ..lowest import rtoken as rtoken_lmodel


def validate_insert(account_id):
    '''
    Validate data use to insert

    :param bson.ObjectId account_id: Account identity corresponding
        with refresh token
    :raise jsonschema.exceptions.ValidationError: On data is invalid
    '''

    jsonschema.validate(account_id, rtoken_lmodel.attr_account_id)



import copy

from ..repo import scheduler as rmodel_sched


#: Show scheduler is cloned by user or not
attr_cloned = {'type': 'boolean'}

#: Show scheduler is updated or not
attr_updated = {'type': 'boolean'}

find_on_cview = copy.deepcopy(rmodel_sched.find)
find_on_cview['required'].extend(['_updated'])
find_on_cview['properties'].update({
    '_updated': attr_updated
})

find_on_uview = copy.deepcopy(rmodel_sched.find)
find_on_uview['required'].extend(['_cloned', '_updated'])
find_on_uview['properties'].update({
    '_cloned': attr_cloned,
    '_updated': attr_updated
})

search_on_cview = copy.deepcopy(rmodel_sched.search)
search_on_cview['items']['required'].extend(['_updated'])
search_on_cview['items']['properties'].update({
    '_updated': attr_updated
})

search_on_uview = copy.deepcopy(rmodel_sched.search)
search_on_uview['items']['required'].extend(['_cloned', '_updated'])
search_on_uview['items']['properties'].update({
    '_cloned': attr_cloned,
    '_updated': attr_updated
})



class Selector(object):
    '''
    Contains informations which is use to search item from resource

    :param keyword: string:
        Search item which is match with keyword
    :type keyword: string
    :param page_index:
        Index of page, start from 1
    :type page_index: int
    :param page_size:
        Max items in one page
    :type page_size: int
    '''
    def __init__(self, keyword=None, page_index=1, page_size=16):
        self.keyword = keyword          #: Get and set keyword
        self.page_index = page_index    #: Get and set page_index
        self.page_size = page_size      #: Get and set page_size

    @property
    def skip(self):
        '''
        :return: Number of item to skip to start of first item to get
        :rtype: int
        '''
        return (self.page_index - 1) * self.page_size


def parse_selector(req):
    '''
    Parse request and return selector. If parameters is not specify in url,
    set it to default

    :param falcon.Request req: Falcon request object
    :return: Instance of selector
    :rtype: gwisp.Selector
    '''

    selector = Selector()

    if 'keyword' in req.params:
        selector.keyword = req.params['keyword']
    if 'page_size' in req.params:
        selector.page_size = int(req.params['page_size'])
    if 'page_index' in req.params:
        selector.page_index = int(req.params['page_index'])

    return selector


class AuthInfo(object):
    '''
    Contains authenticating and account information after authenticate

    :param bson.objectid.ObjectId acc_id: Identity of account
    :param dict token: Token attribute
    :param dict acc: Account attribute
    '''

    def __init__(self, acc_id, token, acc=None):
        self._account_id = acc_id
        self._token = token
        self._account = acc

    @property
    def account_id(self):
        '''
        :return: Account identity
        :rtype: bson.objectid.ObjectId
        '''
        return self._account_id

    @property
    def token(self):
        '''
        :return: Token attribute
        :rtype: dict
        '''
        return self._token

    @property
    def account(self):
        '''
        :return: Account attribute
        :rtype: dict
        '''
        return self._account



import copy
import jsonschema

from ..lowest import account as account_lmodel
from ..engine import account as account_emodel


#: Model use to validate insert
insert = copy.deepcopy(account_emodel.insert)

#: Model use to validate update
update = copy.deepcopy(account_emodel.update)


def validate_insert(item):
    '''
    Validate data use to insert

    :param dict item: Data use to insert
    :raise jsonschema.exceptions.ValidationError: On data is invalid
    '''

    jsonschema.validate(item, insert)


def validate_update(item):
    '''
    Validate data use to update

    :param dict item: Data use to update
    :raise jsonschema.exceptions.ValidationError: On data is invalid
    '''

    jsonschema.validate(item, update)


def validate_groups(data):
    '''
    Validate groups of account

    :param list data: Groups of account
    :raise jsonschema.exceptions.ValidationError: On data is invalid
    '''

    jsonschema.validate(data, account_lmodel.attr_groups)



from .ess import base_64
from .account import attr_id as account_id


#: Account identity corresponding with password hash
attr_id = account_id

#: Password hash
attr_password_hash = base_64

#: Created date. It will nerver changed
attr_created_date = {
    'type': 'string',
    'format': 'date-time'
}

#: Modified date. It will updated when password changed
attr_modified_date = {
    'type': 'string',
    'format': 'date-time'
}

#: Model store in database
storage = {
    'type': 'object',
    'additionalProperties': False,
    'required': ['_id', 'password_hash', 'created_date', 'modified_date'],
    'properties': {
        '_id': attr_id,
        'password_hash': attr_password_hash,
        'created_date': attr_created_date,
        'modified_date': attr_modified_date
    }
}



from datetime import datetime
from bson.objectid import ObjectId

from ..validator import method
from ..model.lowest import password as lmodel_password
from ..model.engine import password as emodel_password


class PasswordRepo(object):
    '''
    Provide method to store password hashed and related meta data

    :param pymongo.MongoClient db: Instance of connection to mongodb server
    '''

    def __init__(self, db):
        self._db = db
        self._coll = db['password']

    @method(ObjectId, lmodel_password.attr_password_hash)
    def insert_one(self, account_id, password_hash):
        '''
        Insert one password object

        :param bson.objectid.ObjectId account_id: Account identity
            correspond with password
        :param string password_hash: Password hashed
        :raise jsonschema.exceptions.ValidationError: On input data is invalid
        '''

        password = {
            '_id': account_id,
            'password_hash': password_hash,
            'created_date': datetime.now().isoformat(),
            'modified_date': datetime.now().isoformat()
        }

        emodel_password.validate_insert(password)

        self._coll.insert_one(password)

    @method(ObjectId, lmodel_password.attr_password_hash)
    def is_exist(self, account_id, passwd_hash):
        '''
        Check pair <account identity, password> is exist

        :param bson.objectid.ObjectId account_id: Account identity
        :param str passwd_hash: Password hashed
        '''

        filters = {'_id': account_id, 'password_hash': passwd_hash}

        return self._coll.find_one(filters) is not None

    @method(ObjectId, lmodel_password.attr_password_hash)
    def update_one(self, account_id, password_hash):
        '''
        Update one password

        :param bson.objectid.ObjectId account_id: Account identity
            correspond with password
        :param string password_hash: Password hashed
        :raise jsonschema.exceptions.ValidationError: On input data is invalid
        '''

        data = {
            'password_hash': password_hash,
            'modified_date': datetime.now().isoformat()
        }

        emodel_password.validate_update(data)

        self._coll.update_one({'_id': account_id}, {'$set': data})



import json
import falcon
import datetime
from json import JSONEncoder
from bson import ObjectId
from urllib.parse import parse_qsl


class RequireJSON(object):

    def process_request(self, req, resp):
        if not req.client_accepts_json:
            raise falcon.HTTPNotAcceptable(
                'This API only supports responses encoded as JSON.',
                href='http://docs.examples.com/api/json')

        if req.method in ('POST', 'PUT'):
            if req.content_type is None:
                raise falcon.HTTPUnsupportedMediaType(
                    'This API only supports requests encoded as JSON or form.',
                    href='http://docs.examples.com/api/json')

            if 'application/json' in req.content_type:
                return
            if 'application/x-www-form-urlencoded' in req.content_type:
                return

            raise falcon.HTTPUnsupportedMediaType(
                'This API only supports requests encoded as JSON or form.',
                href='http://docs.examples.com/api/json')


class JsonEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, ObjectId):
            return str(obj)
        elif isinstance(obj, datetime.datetime):
            return obj.timestamp()

        return JSONEncoder.default(self, obj)


class FormTranslator(object):
    def process_request(self, req, res):
        if req.content_type is None:
            return
        if 'application/x-www-form-urlencoded' not in req.content_type:
            return

        body = req.stream.read()
        req.context['body'] = dict(parse_qsl(body.decode('utf-8')))

    def process_response(self, req, res, data):
        return


class JsonTranslator(object):
    def __init__(self):
        self._json_encoder = JsonEncoder()

    def process_request(self, req, res):
        if req.content_type is None:
            return
        if 'application/json' not in req.content_type:
            return

        if req.content_length in (None, 0):
            req.context['body'] = None
            return

        body = req.stream.read()
        if not body:
            raise falcon.HTTPBadRequest('Empty request body',
                                        'A valid JSON document is required.')

        try:
            req.context['body'] = json.loads(body.decode('utf-8'))

        except (ValueError, UnicodeDecodeError):
            raise falcon.HTTPError(falcon.HTTP_753,
                                   'Malformed JSON',
                                   'Could not decode the request body. The '
                                   'JSON was incorrect or not encoded as '
                                   'UTF-8.')

    def process_response(self, req, res, data):
        res.body = self._json_encoder.encode(res.body)



import copy

from .ess import account_id
from .scheduler import storage as sched_storage

#: Account identity who is used scheduler
attr_cloner = account_id

#: Cloned date
attr_cloned_date = {
    'type': 'integer'
}

#: Model store in database
storage = copy.deepcopy(sched_storage)
storage['required'].extend(['_cloner', '_cloned_date'])
storage['properties'].update({
    '_cloner': attr_cloner,
    '_cloned_date': attr_cloned_date
})



from falcon import HTTP_409, HTTP_401, HTTP_403, HTTP_400, \
                   HTTPNotFound


def dupkey_handle(e, req, res, params):
    res.status = HTTP_409
    res.body = {
        'message': str(e)
    }


def validation_handle(e, req, res, params):
    res.status = HTTP_400
    res.body = {
        'message': str(e)
    }


def jwt_decode_handle(e, req, res, params):
    res.status = HTTP_401
    res.body = {
        'message': str(e)
    }


def jwt_sign_handle(e, req, res, params):
    res.status = HTTP_401
    res.body = {
        'message': str(e)
    }


def unauthorized_handle(e, req, res, params):
    res.status = HTTP_401
    res.body = {
        'message': str(e)
    }


def forbidden_handle(e, req, res, params):
    res.status = HTTP_403
    res.body = {
        'message': str(e)
    }


def not_exist_handle(e, req, res, params):
    raise HTTPNotFound(description=str(e))



import re
from math import ceil
from time import time
from bson import ObjectId

from .scheduler import SchedulerRepo

from ..error import NotExistError
from ..util import Selector
from ..validator import method
from ..model.engine import cscheduler as emodel_csched


search_by_author_fields = {
    'tags': 0, 'tasks': 0, 'notes': 0
}
search_by_cloner_fields = {
    'tags': 0, 'tasks': 0, 'notes': 0, '_cloner': 0
}


class CSchedulerRepo(object):
    '''
    Provide methods to manage cloned schedulers. It will help user use
    schedulers without affected from updating by author of schedulers

    :param pymongo.MongoClient db: Instance of connection to mongdb server
    '''
    def __init__(self, db):
        self._sched_repo = SchedulerRepo(db)
        self._sched_coll = db['scheduler']
        self._csched_coll = db['cscheduler']
        self._acc_coll = db['account']

    @method(ObjectId, ObjectId)
    def clone_one(self, cloner, sched_id):
        '''
        Create an cloned scheduler belong account if it is not early cloned.
        Replace data of scheduler if it is early cloned

        :param bson.objectid.ObjectId cloner: Account identity
            which is clone scheduler
        :param bson.objectid.ObjectId sched_id: Scheduler identity to clone
        '''

        # Find scheduler to clone
        real_sched = self._sched_repo.find_by_id(sched_id)
        if real_sched is None:
            raise NotExistError('Scheduler is not exist')

        # padding data
        real_sched['_cloner'] = cloner
        real_sched['_cloned_date'] = int(time())

        # validate data
        emodel_csched.validate_insert(real_sched)

        # insert or update scheduler depend on it is early exists or not
        if (self._is_exist(cloner, sched_id)):
            self._csched_coll.update_one(
                {'_id': sched_id, '_cloner': cloner},
                {'$set': real_sched}
            )
        else:
            self._csched_coll.insert_one(real_sched)

    @method(ObjectId, Selector)
    def search_on_uview(self, observer, selector=Selector()):
        '''
        Search schedulers on user view correspond with observer.
        On that view, observer see that schedulers is cloned by they or not.
        And if schedulers is cloned by they, it is updated with real schedulers
        or not.
        Two field are added to each schedulers: **_cloned** and **_updated**

        :param bson.objectid.ObjectId observer: Identity of account
        :param gwisp.Selector selector: Keyword and pagging information
        :return: List of scheduler match with selector. None on not found
        :rtype: list
        :rtype: None
        '''

        # search real schedulers
        real_items = self._sched_repo.search(selector)
        if real_items is None:
            return None

        # search cloned schedulers of account
        real_ids = [i['_id'] for i in real_items]
        cloned_cusor = self._csched_coll.find({
            '_cloner': observer,
            '_id': {'$in': real_ids}
        })
        cloned_items = list(cloned_cusor)

        # compare real and cloned schedulers to mark '_cloned' and '_updated'
        cloned_ids = [i['_id'] for i in cloned_items]
        cloned_index = {i['_id']: i for i in cloned_items}
        for item in real_items:
            # scheduler is not cloned
            if item['_id'] not in cloned_ids:
                item['_cloned'] = False
                item['_updated'] = False
                continue

            # scheduler is cloned
            item['_cloned'] = True
            cloned_item = cloned_index[item['_id']]
            if item['modified'] > cloned_item['_cloned_date']:
                item['_updated'] = False
            else:
                item['_updated'] = True

        return real_items

    @method(ObjectId, Selector)
    def search_on_cview(self, cloner, selector=Selector()):
        '''
        Search cloned schedulers on cloner view itself. On that view,
        cloner see that cloned schedulers is updated with real schedulers
        or not. One field added: '_updated'

        :param bson.objectid.ObjectId cloner: Account identity
        :param gwisp.Selector selector: Search information
        '''

        # search schedulers is cloned by account
        filters = {'$and': [{'_cloner': cloner}]}
        if selector.keyword is not None:
            filters['$and'].append(self._gen_filter(selector.keyword))

        cloned_cusor = self._csched_coll.find(
            filters,
            search_by_cloner_fields,
            skip=selector.skip,
            limit=selector.page_size
        )
        cloned_items = list(cloned_cusor)

        # search real schedulers correspond with cloned schedulers
        cloned_ids = [i['_id'] for i in cloned_items]
        real_items = self._sched_coll.find(
            {'_id': {'$in': cloned_ids}},
            {'notes': 0, 'tags': 0, 'tasks': 0, 'author': 0, 'num_user': 0}
        )

        # compare cloned schedulers with real schedulers to mar '_updated'
        real_indexs = {i['_id']: i['modified'] for i in real_items}
        for item in cloned_items:
            if real_indexs[item['_id']] > item['_cloned_date']:
                item['_updated'] = False
            else:
                item['_updated'] = True

            # remove data
            del item['_cloned_date']

        return cloned_items

    @method(ObjectId, ObjectId)
    def find_by_id_on_uview(self, observer, sched_id):
        '''
        Find one scheduler on user view correspond with observer.
        On that view, observer see that scheduler is cloned by they or not.
        And if scheduler is cloned by they, it is updated with real scheduler
        or not. Two fields are added: **_cloned** and **_updated**

        :param bson.objectid.ObjectId observer: Account identity
        :param bson.objectid.ObjectId sched_id: Scheduler identity
        :return dict: Real scheduler data on found. None on not found
        :rtype: dict
        :rtype: None
        '''

        # search real scheduler
        real_item = self._sched_repo.find_by_id(sched_id)
        if real_item is None:
            return None

        # search cloned scheduler
        cloned_item = self._csched_coll.find_one({
            '_id': sched_id,
            '_cloner': observer
        })

        # scheduler is not cloned
        if cloned_item is None:
            real_item['_cloned'] = False
            real_item['_updated'] = False
            return real_item

        # scheduler is cloned, mark '_updated' field
        real_item['_cloned'] = True
        if real_item['modified'] > cloned_item['_cloned_date']:
            real_item['_updated'] = False
        else:
            real_item['_updated'] = True

        return real_item

    @method(ObjectId, ObjectId)
    def find_by_id_on_cview(self, cloner, sched_id):
        '''
        Find cloned scheduler on cloner view itself. On that view, cloner see
        that cloned scheduler is updated with real scheduler or not.
        One field added: ‘_updated’

        :param bson.objectid.ObjectId cloner: Cloner identity
        :param bson.objectid.ObjectId sched_id: Scheduler identity
        :return: Cloned scheduler data on found. None on not found
        :rtype: dict
        :rtype: None
        '''

        # find real scheduler
        real_item = self._sched_repo.find_by_id(sched_id)
        if real_item is None:
            return None

        # find cloned scheduler
        cloned_item = self._csched_coll.find_one(
            {'_id': sched_id, '_cloner': cloner}
        )
        if cloned_item is None:
            return None

        # compare real and cloned scheduler to mark '_updated' field
        if real_item['modified'] > cloned_item['_cloned_date']:
            cloned_item['_updated'] = False
        else:
            cloned_item['_updated'] = True

        # remove data
        del cloned_item['_cloner']
        del cloned_item['_cloned_date']

        return cloned_item

    @method(ObjectId, ObjectId)
    def use_scheduler(self, account_id, sched_id):
        '''
        Make account with account_id use scheduler with sched_id.
        If scheduler is not cloned, clone it.
        Then copy cloned version to account.scheduler

        :param bson.objectid.ObjectId account_id: Account identity
        :param bson.objectid.ObjectId sched_id: Scheduler identity
        '''

        # find cloned scheduler
        cloned_item = self._csched_coll.find_one({
            '_cloner': account_id,
            '_id': sched_id
        })

        # if scheduler is not cloned, clone it
        if cloned_item is None:
            cloned_item = self._sched_repo.find_by_id(sched_id)
            if cloned_item is None:
                raise NotExistError('Scheduler is not exists')

            cloned_item['_cloner'] = account_id
            cloned_item['_cloned_date'] = int(time())

            self._csched_coll.insert_one(cloned_item)

        # copy cloned scheduler to account.scheduler
        del cloned_item['_cloner']
        self._acc_coll.update_one(
            {'_id': account_id},
            {'$set': {'scheduler': cloned_item}}
        )

    @method(ObjectId)
    def update_all_cloned(self, cloner):
        '''
        Update all of cloned schedulers is cloned by cloner with real
        schedulers

        :param bson.objectid.ObjectId cloner: Account identity
        '''

        # find schedulers is using by account
        account = self._acc_coll.find_one(
            {'_id': cloner},
            {'email': 0, 'avatar': 0, 'groups': 0, 'subject': 0}
        )
        if account is None:
            raise NotExistError('Account is not exist')

        # prepare for cloned schedulers
        number_item = self._csched_coll.count({'_cloner': cloner})
        page_size = 16
        number_of_page = ceil(number_item / page_size)

        # clone each page of schedulers
        for page_index in range(number_of_page):
            # query page of cloned schedulers
            cloned_cusor = self._csched_coll.find(
                {'_cloner': cloner},
                {
                    'tags': 0, 'tasks': 0, 'notes': 0, 'author': 0,
                    'modified': 0
                },
                skip=page_index*page_size,
                limit=page_size
            )
            cloned_items = list(cloned_cusor)

            # query page of real schedulers
            cloned_ids = [i['_id'] for i in cloned_items]
            real_cusor = self._sched_coll.find(
                {'_id': {'$in': cloned_ids}}
            )
            real_items = list(real_cusor)

            # find scheduler is using by account then update
            # and padding meta data for each real schedulers
            for item in real_items:
                item['_cloned_date'] = int(time())

                if 'scheduler' in account:
                    if item['_id'] == account['scheduler']['_id']:
                        self._acc_coll.update_one(
                            {'_id': cloner},
                            {'$set': {'scheduler': item}}
                        )

                item['_cloner'] = cloner

            # remove all of cloned schedulers
            self._csched_coll.remove({'_id': {'$in': cloned_ids}})

            # create new cloned schedulers
            self._csched_coll.insert(real_items)

    @method(ObjectId, ObjectId)
    def update_one_cloned(self, cloner, sched_id):
        '''
        Update one cloned scheduler.
        If scheduler is using by cloner, update it

        :param bson.objectid.ObjectId cloner: Account identity
        :param bson.objectid.ObjectId sched_id: Scheduler identity
        '''

        # find real scheduler
        real_item = self._sched_coll.find_one({'_id': sched_id})
        if real_item is None:
            raise NotExistError('Not found scheduler')

        # find account
        account = self._acc_coll.find_one({'_id': cloner})
        if account is None:
            raise NotExistError('Account is not exist')

        # find cloned scheduler
        cloned_item = self._csched_coll.find_one({
            '_id': sched_id,
            '_cloner': cloner
        })
        if cloned_item is None:
            raise NotExistError('Scheduler is not cloned')

        # padding meta data to scheduler
        real_item['_cloned_date'] = int(time())

        # update using scheduler if match
        if 'scheduler' in account and account['scheduler']['_id'] == sched_id:
                self._acc_coll.update_one(
                    {'_id': cloner},
                    {'$set': {'scheduler': real_item}}
                )

        # update cloned scheduler
        self._csched_coll.update_one(
            {'_id': sched_id, '_cloner': cloner},
            {'$set': real_item}
        )

    @method(ObjectId, ObjectId)
    def remove_one_cloned(self, cloner, sched_id):
        '''
        Remove cloned scheduler from account.
        If cloned scheduler is using by account, remove it

        :param bson.objectid.ObjectId cloner: Account identity
        :param bson.objectid.ObjectId sched_id: Scheduler identity
        '''

        # remove cloned scheduler
        result = self._csched_coll.delete_one({
            '_id': sched_id,
            '_cloner': cloner
        })
        if result.deleted_count == 0:
            raise NotExistError('Scheduler is not exist')

        # remove using scheduler if match
        account = self._acc_coll.find_one({'_id': cloner})
        if account is None:
            raise NotExistError('Account is not exist')
        if 'scheduler' in account and account['scheduler']['_id'] == sched_id:
            self._acc_coll.update_one(
                {'_id': cloner},
                {'$unset': {'scheduler': None}}
            )

    @method(ObjectId, ObjectId)
    def _is_exist(self, account_id, sched_id):
        '''
        Check scheduler is cloned by account

        :param bson.objectid.ObjectId account_id: Account identity
        :param bson.objectid.ObjectId sched_id: Scheduler identity
        :return: True on scheduler is cloned by account
        :return: False on scheduelr is not cloned by account
        :rtype: bool
        '''

        key = {'account_id': account_id, '_id': sched_id}
        return self._csched_coll.find_one(key) is not None

    def _gen_filter(self, keyword):
        '''
        Generate query filter of mongodb

        :param str keyword: Keyword to search
        :return: Filter
        :rtype: dict
        '''
        return {'$or': [
            {'name': {'$regex': re.compile(keyword)}},
            {'tags': {
                '$elemMatch': {
                    '$regex': re.compile(keyword)
                }
            }}
        ]}



import re
from time import time
from bson.objectid import ObjectId

from ..error import NotExistError
from ..util import Selector
from ..validator import method
from ..model.repo import scheduler as rmodel_sched
from ..model.engine import scheduler as emodel_sched


search_fields = {'notes': 0, 'tasks': 0, 'tags': 0}
search_by_author_fields = {'notes': 0, 'tasks': 0, 'tags': 0}
verify_fields = {
    '_id': 0,
    'name': 0,
    'tags': 0,
    'tasks': 0,
    'notes': 0
}


class SchedulerRepo(object):
    '''
    Provide method to manage schedulers in database

    :param pymongo.MongoClient db: Instance of MongoClient
    '''
    def __init__(self, db):
        self._db = db
        self._coll = db['scheduler']
        self._acc_coll = db['account']

    @method(rmodel_sched.insert)
    def insert_one(self, item):
        '''
        Insert one scheduler

        :param dict item: Scheduler to insert
        :return: Identity of new scheduler
        :rtype: bson.objectid.ObjectId
        :raise jsonschema.exceptions.ValidationError: On item is invalid
        '''

        # padding data
        item['num_user'] = 0
        item['modified'] = int(time())

        emodel_sched.validate_insert(item)

        # save to database
        return self._coll.insert_one(item).inserted_id

    @method(rmodel_sched.insert_many)
    def insert_many(self, items):
        '''
        Insert multi schedulers

        :param list items: List of schedulers
        :return: List of identity of new scheduler
        :rtype: list
        :raise jsonschema.exceptions.ValidationError: On item is invalid
        '''

        # padding data
        for item in items:
            item['num_user'] = 0
            item['modified'] = int(time())

        emodel_sched.validate_insert(items)

        # save to database
        return self._coll.insert_many(items).inserted_ids

    @method(ObjectId)
    def find_by_id(self, id):
        '''
        Find scheduler by identity

        :param bson.objectid.ObjectId id: Identity of scheduler
        :return: Scheduler dictionary on found
        :return: None of not found
        :rtype: dict
        '''

        return self._coll.find_one({'_id': id})

    @method(Selector)
    def search(self, selector=Selector()):
        '''
        Search scheduler match with selector

        :param gwisp.Selector selector: Contains specification to search
        :return: List of dictionary contains essential information of scheduler
              on found
        :return: List with none element not found
        :rtype: list
        '''

        filters = {}
        if selector.keyword is not None:
            filters = self._gen_filter(selector.keyword)

        cursor = self._coll.find(filters,
                                 search_fields,
                                 skip=selector.skip,
                                 limit=selector.page_size)

        if cursor.count(with_limit_and_skip=True) == 0:
            return None
        return list(cursor)

    @method(ObjectId, Selector)
    def search_by_author(self, author_id, selector=Selector()):
        '''
        Search by author identity and match with selector

        :param bson.objectid.ObjectId author_id: Author identity
        :param gwisp.Selector selector: Contains specification to search
        :return: List of dictionary contains essential information of scheduler
              on found
        :return: List with none element not found
        :rtype: list
        '''

        filters = {'$and': [{'author._id': author_id}]}
        if selector.keyword is not None:
            filters['$and'].append(self._gen_filter(selector.keyword))

        cursor = self._coll.find(
            filters,
            search_by_author_fields,
            skip=selector.skip,
            limit=selector.page_size
        )

        if cursor.count(with_limit_and_skip=True) == 0:
            return None
        return list(cursor)

    @method(ObjectId, rmodel_sched.update)
    def update_one(self, id, item):
        '''
        Update one scheduler

        :param bson.ObjectId id: Scheduler identity
        :param dict item: Scheduler to update, must contain _id fields
        :raise jsonschema.exceptions.ValidationError: On item is invalid
        '''

        # padding data
        item['modified'] = int(time())

        emodel_sched.validate_update(item)

        result = self._coll.update_one({'_id': id}, {'$set': item})

        if result.matched_count == 0:
            raise NotExistError('Scheduler is not exist')

    @method(ObjectId)
    def delete_one(self, id):
        '''
        Delete one scheduler

        :param bson.objectid.ObjectId id: Identity of scheduler to delete
        :raise gwisp.error.NotExistError: On scheduler is not exist
        '''

        result = self._coll.delete_one({'_id': id})

        if result.deleted_count == 0:
            raise NotExistError('Scheduler is not exist')

    @method(ObjectId)
    def is_exist(self, id):
        '''
        Verify that scheduler is exist

        :param bson.objectid.ObjectId id: Identity of scheduler
        :return: True on scheduler is exist
        :return: False on scheduler is not exist
        :rtype: bool
        '''

        return self._coll.find_one({'_id': id}, verify_fields) is not None

    def count(self):
        '''
        Count number of schedulers

        :return: Number of schedulers in collection
        :rtype: int
        '''

        return self._coll.count()

    def _gen_filter(self, keyword):
        '''
        Generate query filter of mongodb

        :param str keyword: Keyword to search
        :return: Filter
        :rtype: dict
        '''
        return {'$or': [
            {'name': {'$regex': re.compile(keyword)}},
            {'tags': {
                '$elemMatch': {
                    '$regex': re.compile(keyword)
                }
            }}
        ]}



from falcon import HTTPNotFound

from ..repo import SchedulerRepo
from ..util import parse_selector


class SchedulerMeCtl(object):
    '''
    Allow search schedulers by author

    :param pymongo.MongoClient db: Instance of connection to mongodb server
    :param gwisp.Protector protector: Instance of protector
    '''

    def __init__(self, db, protector):
        self._repo = SchedulerRepo(db)
        self._protector = protector

    def on_get(self, req, res):
        auth = self._protector.check(req)
        selector = parse_selector(req)

        res.body = self._repo.search_by_author(auth.account_id, selector)
        if res.body is None:
            raise HTTPNotFound(
                title='404 Not Found',
                description='Not found scheduler'
            )
        else:
            # remove data
            for item in res.body:
                del item['author']



import copy

from .get_scheduler import res_body as get_sched_res_body


res_body = copy.deepcopy(get_sched_res_body)
res_body['items']['required'].remove('author')
res_body['items']['properties'].pop('author')
res = {'body': res_body}



res_body = {
    'type': 'object',
    'additionalProperties': False,
    'required': [
        'name', 'notes', 'version', 'source', 'doc',
        'num_user', 'num_sched'
    ],
    'properties': {
        'name': {'type': 'string'},
        'notes': {'type': 'string'},
        'version': {
            'type': 'string',
            'pattern': '^[0-9]+\.[0-9]+\.[0-9]+$',
        },
        'source': {'type': 'string'},
        'doc': {'type': 'string'},
        'num_user': {'type': 'integer'},
        'num_sched': {'type': 'integer'}
    }
}

res = {'body': res_body}



from ..model.setting import validate_insert
from ..error import NotAllowError


class SettingRepo(object):
    '''
    Setting Repository

    :param pymongo.MongoClient db: Instance connection to mongo server
    '''

    def __init__(self, db):
        self._db = db
        self._coll = db['setting']

    def insert_one(self, item):
        '''
        Insert one setting into database

        :pram dict item: Setting item
        '''

        validate_insert(item)

        if not self.is_empty():
            raise NotAllowError('Allow only item in collection')

        self._coll.insert_one(item)

    def is_empty(self):
        '''
        Check collection is empty

        :return: True on empty, False on otherwise
        :rtype: bool
        '''

        return self._coll.find_one() is None



#!/usr/bin/env python

import platform
import sys
import shutil
from os import path, makedirs
from urllib import request
from subprocess import Popen
from setuptools import setup, find_packages, Command


root = path.realpath(path.dirname(__file__))


class InitCommand(Command):
    user_options = []
    description = 'Install essential package to manager python'

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        # install pip - python package manager
        url = 'https://bootstrap.pypa.io/get-pip.py'
        tmp_dir = path.join(root, 'tmp')
        install_file = path.join(tmp_dir, 'install-pip.py')
        if not path.isdir(tmp_dir):
            makedirs(tmp_dir)
        request.urlretrieve(url, install_file)
        if Popen([sys.executable, install_file]).wait() != 0:
            sys.exit(1)

        # install virtualenv package
        cmd = ['pip', 'install', 'virtualenv']
        if platform.os.name == 'posix':
            cmd.insert(0, 'sudo')
        if Popen(cmd).wait() != 0:
            sys.exit(1)

        # create virtual environment
        if Popen(['virtualenv', 'venv']).wait() != 0:
            sys.exit(1)


class DevCommand(Command):
    user_options = []
    description = 'Install packages for development'

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        # install develop packages
        dev_file = path.join(root, 'dev-requires.txt')
        if Popen(['pip', 'install', '-r', dev_file]).wait() != 0:
            sys.exit(1)

        # install runtime dependency packages
        if Popen(['pip', 'install', '-e', '.']).wait() != 0:
            sys.exit(1)


class CleanCommand(Command):
    user_options = []
    description = 'Clean all of build result'

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        shutil.rmtree(path.join(root, '.cache'), True)
        shutil.rmtree(path.join(root, 'dist'), True)
        shutil.rmtree(path.join(root, 'build'), True)
        shutil.rmtree(path.join(root, 'dest'), True)
        shutil.rmtree(path.join(root, 'gwisp.egg-info'), True)
        shutil.rmtree(path.join(root, '.eggs'), True)
        shutil.rmtree(path.join(root, 'tmp'), True)
        shutil.rmtree(path.join(root, 'temp'), True)


class CleanVenvCommand(Command):
    user_options = []
    description = 'Clean virtual environment'

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        venv_dir = path.join(root, 'venv')
        shutil.rmtree(venv_dir, True)


setup(
    name='gwisp',
    version='0.0.24',
    description='http programing interface of scheduler storage',
    long_description='see more on home page',
    keywords='schedueler time storage api',
    author='kevin leptons',
    author_email='kevin.leptons@gmail.com',
    url='https://github.com/kevin-leptons/gwisp',
    download_url='https://github.com/kevin-leptons/gwisp',
    install_requires=[
        'falcon', 'pymongo', 'google-api-python-client',
        'pyjwt',
        'click', 'jsonschema', 'waitress'
    ],
    packages=find_packages(exclude=['script', 'test']),
    entry_points={
        'console_scripts': [
            'gwisp=gwisp.cli:cli',
        ]
    },
    cmdclass={
        'init': InitCommand,
        'clean': CleanCommand,
        'cleanvenv': CleanVenvCommand,
        'dev': DevCommand,
    }
)



from pymongo import ASCENDING


class CollectionSpec(object):
    '''
    Contains information of index of mongodb collection

    :param str name: Name of collection
    '''
    def __init__(self, name):
        self._name = name
        self._indexes = {}

    @property
    def name(self):
        '''
        :return: Name of collection
        :rtype: str
        '''

        return self._name

    @property
    def indexes(self):
        '''
        :return: Dictionary contains pair ``str``-> ``bool``.
            Key is index name, value is dictionary contains
            ``key`` and ``unique`` key
        :rtype: dict
        '''

        return self._indexes

    def add_index(self, key, unique):
        '''
        Add an index to collection

        :param list key: List of string make a key
        :param bool unique: Mark key is unique
        '''

        self._indexes['_'.join(key)] = {
            'key': [(k, ASCENDING) for k in key],
            'unique': unique
        }


#: Password collection specification
password = CollectionSpec('password')

#: Account collection specification
account = CollectionSpec('account')
account.add_index(['name'], True)
account.add_index(['name', 'subject'], True)

#: Refresh token specifications
rtoken = CollectionSpec('rtoken')

#: Scheduler collection specification
scheduler = CollectionSpec('scheduler')
scheduler.add_index(['name'], True)

#: Cloned scheduler
cscheduler = CollectionSpec('cscheduler')
cscheduler.add_index(['_id', '_cloner'], True)

#: List of collection specifications in database
specs = [password, account, rtoken, scheduler, cscheduler]



from .root import RootCtl
from .oauth import OauthCtl, LoginUrlCtl, LoginCtl, MeCtl
from .scheduler import SchedulerCtl, SchedulerItemCtl
from .scheduler_user import SchedulerUserCtl, SchedulerUserItemCtl
from .scheduler_me import SchedulerMeCtl
from .me_target import MeTargetCtl, MeTargetItemCtl
from .me_scheduler import MeSchedulerCtl, MeSchedulerItemCtl



import pkg_resources

from ..repo import AccountRepo, SchedulerRepo


class RootCtl(object):
    '''
    Information of service. Read only resource
    '''

    def __init__(self, db):
        '''
        :param pymongo.MongoClient db: Instance of connection to mongodb server
        '''

        self._acc_repo = AccountRepo(db)
        self._sched_repo = SchedulerRepo(db)

    def on_get(self, req, res):
        '''
        Response: Essential information of service
        '''

        data = {
            'name': 'gwisp service',
            'notes': 'programing interface of scheduler storage on http',
            'version': pkg_resources.get_distribution('gwisp').version,
            'source': 'https://github.com/kevin-leptons/gwisp',
            'doc': 'http://gwisp.readthedocs.io/en/latest/',
            'num_user': self._acc_repo.count(),
            'num_sched': self._sched_repo.count()
        }

        res.body = data



import copy
import jsonschema

from ..lowest import password


#: Model use to validate insert data
insert = copy.deepcopy(password.storage)

#: Model use to validate update data
update = copy.deepcopy(password.storage)
update['required'].remove('_id')
update['required'].remove('created_date')
update['properties'].pop('_id')
update['properties'].pop('created_date')


def validate_insert(item):
    '''
    Validate insert data

    :param dict: Data use to insert
    :raise jsonschema.exceptions.ValidationError: On data is invalid
    '''

    jsonschema.validate(item, insert)


def validate_update(item):
    '''
    Validate update data

    :param dict: Data use to update
    :raise jsonschema.exceptions.ValidationError: On data is invalid
    '''

    jsonschema.validate(item, update)



from .worker import Worker
from waitress import serve


class Service(object):
    '''
    This is higher public api, allow start gwisp service very quickly.
    Setup service by configuration and call ``start()`` method


    :param int port: Port which is service will be serve
    :param str db_url: Url to mongo database server
    :param str client_id: Google client identity, use for oauth-2
    :param str client_secret: Google client secret, use for oauth-2
    :param str redirect_uri: Callback url after confirm google login
    :param str jwt_key: Secret words use to sign token of service
    '''
    def __init__(self, port, db_url,
                 client_id, client_secret, redirect_uri, jwt_key):
        self._port = port
        self._db_url = db_url
        self._client_id = client_id
        self._client_secret = client_secret
        self._redirect_uri = redirect_uri
        self._jwt_key = jwt_key

    def reg_root(self, email, passwd):
        '''
        Register root account. Name is gwisp

        :param str email: Email of root account
        :param str passwd: Password in text plain
        '''

        worker = self._gen_worker()
        worker.reg_root(email, passwd)

        worker.stop()

    def start(self):
        '''
        Use waitress as server to create worker and serve
        '''

        worker = self._gen_worker()
        app = worker.start()

        if self._port > 0:
            serve(app, host='0.0.0.0', port=self._port)
        else:
            serve(app)

    def renew_db(self):
        '''
        Ensure database have correct index and empty data
        '''

        worker = self._gen_worker()
        worker.renew_db()

        worker.stop()

    def setup(self, asset_dir):
        '''
        Drop database if it is exist. Create new database with correct
        indexs. Install sample data

        :param str asset_dir: Path to asset directory.
            See :ref:`api_asset_dir`
        '''

        worker = self._gen_worker()
        worker.setup(asset_dir)

        worker.stop()

    def _gen_worker(self):
        return Worker(
            self._db_url,
            self._client_id, self._client_secret, self._redirect_uri,
            self._jwt_key
        )
